﻿using ScrambleKit.Extensions;
using UnityEngine;

public class Moto : MonoBehaviour
{
    public string DEBUG;

    public float enginePower = 50f;
    public float driftBrakePower = 10f;
    public float rollingFrictionFactor = 0.3f;

    public Transform face;

    public float maxLean = 55f;
    public float faceLeanFactor = 0.7f;
    public float faceLookMax = 8f;
    public float faceScootFactor = 0.2f;

    public float throttle;
    public float steer;
    public float rawSteer;
    public bool boosting;
    public bool drifting;
    private float driftRot;

    private Rigidbody rb;
    void Start() { rb = GetComponent<Rigidbody>(); }
    
    void FixedUpdate()
    {
        ValidateInputs();
        UpdateAnglesForMotion();
        UpdatePhysics();
    }

    private void ValidateInputs()
    {
        throttle = Mathf.Clamp(throttle, 0, 1);
        steer = Mathf.Clamp(steer, -1, 1);

        if (boosting)
            throttle = 1;
    }

    // FEEL OF CONTROLS, LEAN ETC

    void UpdateAnglesForMotion()
    {
        float velocityInfluence = VelocityLeanInfluence();
        float bodyLeanAngle = velocityInfluence * maxLean * -steer;
        float faceLeanAngle = faceLeanFactor * -bodyLeanAngle;

        float faceLookAngle = faceLookMax * steer;
        float faceScoot = faceScootFactor * throttle;

        float bodyFacingAngle = GetAngleForVelocity();

        if (drifting)  // toggling drift was super-jarring on the rotation, so I'm lerping in and out of it's impact now
        {
            driftRot = Mathf.Lerp(driftRot, 60 * rawSteer * velocityInfluence, 0.25f);
        }
        else
        {
            driftRot = Mathf.Lerp(driftRot, 0, 0.25f);
        }
        bodyFacingAngle += driftRot;

        transform.localEulerAngles = new Vector3(
                transform.localEulerAngles.x,
                Mathf.LerpAngle(transform.localEulerAngles.y, bodyFacingAngle, 0.35f),
                Mathf.LerpAngle(transform.localEulerAngles.z, bodyLeanAngle, 0.35f)
            );

        face.localEulerAngles = new Vector3(
                Mathf.LerpAngle(face.localEulerAngles.x, - Mathf.Abs(faceLookAngle) * 1.2f * velocityInfluence, 0.5f),
                Mathf.LerpAngle(face.localEulerAngles.y, faceLookAngle, 0.35f),
                Mathf.LerpAngle(face.localEulerAngles.z, faceLeanAngle, 0.35f)
            );

        face.localPosition = Vector3.Lerp(face.localPosition, new Vector3(
                face.localPosition.x,
                face.localPosition.y,
                faceScoot
                ),
                0.35f
             );
    }

    private float VelocityLeanInfluence()
    {
        return Mathf.Min(1f, rb.velocity.ToHoriz().magnitude / 24f);
    }

    private float GetAngleForVelocity()
    {
        Vector3 horizVel = rb.velocity.ToHoriz();
        if (horizVel.magnitude == 0)
            return transform.localEulerAngles.y;  // existing angle should remain unchanged

        Vector2 delta2 = new Vector2(horizVel.x, horizVel.z);
        float angle = -delta2.AngleFacing();

        return angle;
    }

    // DRIVING PHYSICS

    void UpdatePhysics()
    {
        Vector3 totalForce = Vector3.zero;
        totalForce += GetSlideResistance();

        if (drifting)
            totalForce *= 0.15f;

        totalForce += GetPower();
        totalForce += GetBraking();
        totalForce += GetRollingResistance();

        //totalForce = HandleTraction(totalForce);

        Vector3 force = totalForce * rb.mass;
        rb.AddForce(force);
    }

    private Vector3 GetSlideResistance()
    {
        // use dot to project forward direction vs angle of steering
        // take the part out of alignment with the wheel and "resist" it

        // I think I will treat this as a "2.5D" system, by projecting it into a 2D world

        Vector2 vel = new Vector2(rb.velocity.x, rb.velocity.z);
        //Vector3 forward25d = transform.forward.ToHoriz().normalized;
        //Vector2 forward = new Vector2(forward25d.x, forward25d.z);
        Vector3 forward25d = transform.TransformVector(new Vector3(steer, 0, 1)).normalized;
        Vector2 forward = new Vector2(forward25d.x, forward25d.z);

        float dot = Vector2.Dot(forward, vel);
        Vector2 projection = dot * forward;  // the part of vel that is okay b/c of rolling
        Vector2 stick = -(vel - projection); // force from fighting the part of vel that's sideways for the tire

        Vector3 stick25d = new Vector3(stick.x, 0, stick.y);

        Vector3 giveback = .6f * stick25d.magnitude * forward25d.normalized;  // turning slows TOO MUCH feelsbad

        Vector3 totalSteer = stick25d + giveback;

        Debug.DrawLine(transform.position, transform.position + 6 * forward25d, Color.white);
        Debug.DrawLine(transform.position, transform.position + rb.velocity.ToHoriz(), Color.cyan);
        Debug.DrawLine(transform.position, transform.position + stick25d, Color.magenta);

        // Tinkering magic numbers to make it feel good to turn fast at low speed, and vice versa:
        float framesToTake = Mathf.Max(4, 0.4f * rb.velocity.ToHoriz().magnitude);
        return totalSteer / (framesToTake * Time.fixedDeltaTime);
    }

    private Vector3 GetPower()
    {
        Vector3 forward25d = transform.forward.ToHoriz().normalized;
        Vector3 power = forward25d * enginePower * throttle;
        Debug.DrawLine(transform.position, transform.position + power, Color.green);

        return power;
    }

    private Vector3 GetBraking()
    {
        if (!drifting)
            return Vector3.zero;

        Vector3 braking = -rb.velocity.ToHoriz().normalized * Mathf.Min(rb.velocity.ToHoriz().magnitude, driftBrakePower);
        Debug.DrawLine(transform.position, transform.position + braking, Color.red);
        return braking;
    }

    private Vector3 GetRollingResistance()
    {
        if (throttle == 0)
            return Vector3.zero;

        Vector3 braking = -rb.velocity.ToHoriz() * rollingFrictionFactor;
        Debug.DrawLine(transform.position, transform.position + braking, Color.yellow);
        return braking;
    }
}
